@@ -1,4 +1,4 @@
-// Copyright 2019, 2020 Parity Technologies
+// Copyright 2019, 2021 Parity Technologies
 //
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
@@ -25,11 +25,11 @@
 //! expected to save roughly (n - 1) hashes in size where n is the number of nodes in the partial
 //! trie.
 
-use hash_db::HashDB;
+use hash_db::{HashDB, Prefix};
 use crate::{
 	CError, ChildReference, DBValue, NibbleVec, NodeCodec, Result,
 	TrieHash, TrieError, TrieDB, TrieDBNodeIterator, TrieLayout,
-	nibble_ops::NIBBLE_LENGTH, node::{Node, NodeHandle, NodeHandlePlan, NodePlan, OwnedNode},
+	nibble_ops::NIBBLE_LENGTH, node::{Node, NodeHandle, NodeHandlePlan, NodePlan, ValuePlan, OwnedNode},
 };
 use crate::rstd::{
 	boxed::Box, convert::TryInto, marker::PhantomData, rc::Rc, result, vec, vec::Vec,
@@ -38,12 +38,15 @@ use crate::rstd::{
 struct EncoderStackEntry<C: NodeCodec> {
 	/// The prefix is the nibble path to the node in the trie.
 	prefix: NibbleVec,
+	/// Node in memory content.
 	node: Rc<OwnedNode<DBValue>>,
 	/// The next entry in the stack is a child of the preceding entry at this index. For branch
 	/// nodes, the index is in [0, NIBBLE_LENGTH] and for extension nodes, the index is in [0, 1].
 	child_index: usize,
 	/// Flags indicating whether each child is omitted in the encoded node.
 	omit_children: Vec<bool>,
+	/// Skip value if value node is after.
+	omit_value: bool,
 	/// The encoding of the subtrie nodes rooted at this entry, which is built up in
 	/// `encode_compact`.
 	output_index: usize,
@@ -95,9 +98,20 @@ impl<C: NodeCodec> EncoderStackEntry<C> {
 	}
 
 	/// Generates the encoding of the subtrie rooted at this entry.
-	fn encode_node(&self) -> Result<Vec<u8>, C::HashOut, C::Error> {
+	fn encode_node(&mut self) -> Result<Vec<u8>, C::HashOut, C::Error> {
 		let node_data = self.node.data();
-		Ok(match self.node.node_plan() {
+		let mut modified_node_plan;
+		let node_plan = if self.omit_value {
+			modified_node_plan = self.node.node_plan().clone();
+			if let Some(value) = modified_node_plan.value_plan_mut() {
+				// 0 length value.
+				*value = ValuePlan::Inline(0 .. 0);
+			}
+			&modified_node_plan
+		} else {
+			self.node.node_plan()
+		};
+		let mut encoded = match node_plan {
 			NodePlan::Empty | NodePlan::Leaf { .. } => node_data.to_vec(),
 			NodePlan::Extension { partial, child: _ } => {
 				if !self.omit_children[0] {
@@ -111,7 +125,7 @@ impl<C: NodeCodec> EncoderStackEntry<C> {
 			NodePlan::Branch { value, children } => {
 				C::branch_node(
 					Self::branch_children(node_data, &children, &self.omit_children)?.iter(),
-					value.clone().map(|range| &node_data[range])
+					value.as_ref().map(|v| v.build(node_data)),
 				)
 			}
 			NodePlan::NibbledBranch { partial, value, children } => {
@@ -120,10 +134,19 @@ impl<C: NodeCodec> EncoderStackEntry<C> {
 					partial.right_iter(),
 					partial.len(),
 					Self::branch_children(node_data, &children, &self.omit_children)?.iter(),
-					value.clone().map(|range| &node_data[range])
+					value.as_ref().map(|v| v.build(node_data)),
 				)
 			}
-		})
+		};
+
+		if self.omit_value {
+			if let Some(header) = C::ESCAPE_HEADER {
+				encoded.insert(0, header);
+			} else {
+				return Err(Box::new(TrieError::InvalidStateRoot(Default::default())));
+			}
+		}
+		Ok(encoded)
 	}
 
 	/// Generate the list of child references for a branch node with certain children omitted.
@@ -158,6 +181,29 @@ impl<C: NodeCodec> EncoderStackEntry<C> {
 	}
 }
 
+/// Dettached value if included does write a reserved header,
+/// followed by node encoded with 0 length value and the value
+/// as a standalone vec.
+fn detached_value<L: TrieLayout>(
+	value: &ValuePlan,
+	node_data: &[u8],
+	node_prefix: Prefix,
+	val_fetcher: &TrieDBNodeIterator<L>,
+) -> Option<Vec<u8>> {
+	let fetched;
+	match value {
+		ValuePlan::Node(hash_plan) => {
+			if let Some(value) = val_fetcher.fetch_value(&node_data[hash_plan.clone()], node_prefix) {
+				fetched = value;
+			} else {
+				return None;
+			}
+		},
+		_ => return None,
+	}
+	Some(fetched)
+}
+
 /// Generates a compact representation of the partial trie stored in the given DB. The encoding
 /// is a vector of mutated trie nodes with those child references omitted. The mutated trie nodes
 /// are listed in pre-order traversal order so that the full nodes can be efficiently
@@ -167,7 +213,7 @@ impl<C: NodeCodec> EncoderStackEntry<C> {
 /// references.
 pub fn encode_compact<L>(db: &TrieDB<L>) -> Result<Vec<Vec<u8>>, TrieHash<L>, CError<L>>
 	where
-		L: TrieLayout
+		L: TrieLayout,
 {
 	let mut output = Vec::new();
 
@@ -179,13 +225,13 @@ pub fn encode_compact<L>(db: &TrieDB<L>) -> Result<Vec<Vec<u8>>, TrieHash<L>, CE
 	// - It yields at least one node.
 	// - The first node yielded is the root node with an empty prefix and is not inline.
 	// - The prefixes yielded are in strictly increasing lexographic order.
-	let iter = TrieDBNodeIterator::new(db)?;
+	let mut iter = TrieDBNodeIterator::new(db)?;
 
 	// Following from the guarantees about TrieDBNodeIterator, we guarantee that after the first
 	// iteration of the loop below, the stack always has at least one entry and the bottom (front)
 	// of the stack is the root node, which is not inline. Furthermore, the iterator is not empty,
 	// so at least one iteration always occurs.
-	for item in iter {
+	while let Some(item) = iter.next() {
 		match item {
 			Ok((prefix, node_hash, node)) => {
 				// Skip inline nodes, as they cannot contain hash references to other nodes by
@@ -216,22 +262,31 @@ pub fn encode_compact<L>(db: &TrieDB<L>) -> Result<Vec<Vec<u8>>, TrieHash<L>, CE
 					}
 				}
 
-				let children_len = match node.node_plan() {
-					NodePlan::Empty | NodePlan::Leaf { .. } => 0,
-					NodePlan::Extension { .. } => 1,
-					NodePlan::Branch { .. } | NodePlan::NibbledBranch { .. } => NIBBLE_LENGTH,
+				let (children_len, detached_value) = match node.node_plan() {
+					NodePlan::Empty => (0, None),
+					NodePlan::Leaf { value, .. } => (0, detached_value(value, node.data(), prefix.as_prefix(), &iter)),
+					NodePlan::Extension { .. } => (1, None),
+					NodePlan::NibbledBranch { value: Some(value), .. }
+					| NodePlan::Branch { value: Some(value), .. } => (NIBBLE_LENGTH, detached_value(value, node.data(), prefix.as_prefix(), &iter)),
+					NodePlan::NibbledBranch { value: None, .. }
+					| NodePlan::Branch { value: None, .. } => (NIBBLE_LENGTH, None),
 				};
+
 				stack.push(EncoderStackEntry {
 					prefix,
 					node,
 					child_index: 0,
 					omit_children: vec![false; children_len],
+					omit_value: detached_value.is_some(),
 					output_index: output.len(),
 					_marker: PhantomData::default(),
 				});
 				// Insert a placeholder into output which will be replaced when this new entry is
 				// popped from the stack.
 				output.push(Vec::new());
+				if let Some(value) = detached_value {
+					output.push(value);
+				}
 			}
 			Err(err) => match *err {
 				// If we hit an IncompleteDatabaseError, just ignore it and continue encoding the
@@ -243,7 +298,7 @@ pub fn encode_compact<L>(db: &TrieDB<L>) -> Result<Vec<Vec<u8>>, TrieHash<L>, CE
 		}
 	}
 
-	while let Some(entry) = stack.pop() {
+	while let Some(mut entry) = stack.pop() {
 		output[entry.output_index] = entry.encode_node()?;
 	}
 
@@ -257,6 +312,8 @@ struct DecoderStackEntry<'a, C: NodeCodec> {
 	child_index: usize,
 	/// The reconstructed child references.
 	children: Vec<Option<ChildReference<C::HashOut>>>,
+	/// A value attached as a node. The node will need to use its hash as value.
+	attached_value: Option<&'a [u8]>,
 	_marker: PhantomData<C>,
 }
 
@@ -347,12 +404,13 @@ impl<'a, C: NodeCodec> DecoderStackEntry<'a, C> {
 	///
 	/// Preconditions:
 	/// - if node is an extension node, then `children[0]` is Some.
-	fn encode_node(self) -> Vec<u8> {
+	fn encode_node(self, attached_hash: Option<&[u8]>) -> Vec<u8> {
+		let attached_hash = attached_hash.map(|h| crate::node::Value::Node(h, None));
 		match self.node {
 			Node::Empty =>
 				C::empty_node().to_vec(),
 			Node::Leaf(partial, value) =>
-				C::leaf_node(partial.right(), value),
+				C::leaf_node(partial.right(), attached_hash.unwrap_or(value)),
 			Node::Extension(partial, _) =>
 				C::extension_node(
 					partial.right_iter(),
@@ -361,13 +419,24 @@ impl<'a, C: NodeCodec> DecoderStackEntry<'a, C> {
 						.expect("required by method precondition; qed"),
 				),
 			Node::Branch(_, value) =>
-				C::branch_node(self.children.into_iter(), value),
+				C::branch_node(
+					self.children.into_iter(),
+					if attached_hash.is_some() {
+						attached_hash
+					} else {
+						value
+					},
+				),
 			Node::NibbledBranch(partial, _, value) =>
 				C::branch_node_nibbled(
 					partial.right_iter(),
 					partial.len(),
 					self.children.iter(),
-					value,
+					if attached_hash.is_some() {
+						attached_hash
+					} else {
+						value
+					},
 				),
 		}
 	}
@@ -385,21 +454,21 @@ impl<'a, C: NodeCodec> DecoderStackEntry<'a, C> {
 //
 /// This function makes the assumption that all child references in an inline trie node are inline
 /// references.
-pub fn decode_compact<L, DB, T>(db: &mut DB, encoded: &[Vec<u8>])
+pub fn decode_compact<L, DB>(db: &mut DB, encoded: &[Vec<u8>])
 	-> Result<(TrieHash<L>, usize), TrieHash<L>, CError<L>>
 	where
 		L: TrieLayout,
-		DB: HashDB<L::Hash, T>,
+		DB: HashDB<L::Hash, DBValue>,
 {
-	decode_compact_from_iter::<L, DB, T, _>(db, encoded.iter().map(Vec::as_slice))
+	decode_compact_from_iter::<L, DB, _>(db, encoded.iter().map(Vec::as_slice))
 }
 
 /// Variant of 'decode_compact' that accept an iterator of encoded nodes as input.
-pub fn decode_compact_from_iter<'a, L, DB, T, I>(db: &mut DB, encoded: I)
+pub fn decode_compact_from_iter<'a, L, DB, I>(db: &mut DB, encoded: I)
 	-> Result<(TrieHash<L>, usize), TrieHash<L>, CError<L>>
 	where
 		L: TrieLayout,
-		DB: HashDB<L::Hash, T>,
+		DB: HashDB<L::Hash, DBValue>,
 		I: IntoIterator<Item = &'a [u8]>,
 {
 	// The stack of nodes through a path in the trie. Each entry is a child node of the preceding
@@ -409,8 +478,15 @@ pub fn decode_compact_from_iter<'a, L, DB, T, I>(db: &mut DB, encoded: I)
 	// The prefix of the next item to be read from the slice of encoded items.
 	let mut prefix = NibbleVec::new();
 
-	for (i, encoded_node) in encoded.into_iter().enumerate() {
-		let node = L::Codec::decode(encoded_node)
+	let mut iter = encoded.into_iter().enumerate();
+	while let Some((i, encoded_node)) = iter.next() {
+		let mut attached_node = 0;
+		if let Some(header) = L::Codec::ESCAPE_HEADER {
+			if encoded_node.starts_with(&[header]) {
+				attached_node = 1;
+			}
+		}
+		let node = L::Codec::decode(&encoded_node[attached_node..])
 			.map_err(|err| Box::new(TrieError::DecoderError(<TrieHash<L>>::default(), err)))?;
 
 		let children_len = match node {
@@ -422,9 +498,19 @@ pub fn decode_compact_from_iter<'a, L, DB, T, I>(db: &mut DB, encoded: I)
 			node,
 			child_index: 0,
 			children: vec![None; children_len],
+			attached_value: None,
 			_marker: PhantomData::default(),
 		};
 
+		if attached_node > 0 {
+			// Read value
+			if let Some((_, fetched_value)) = iter.next() {
+				last_entry.attached_value = Some(fetched_value);
+			} else {
+				return Err(Box::new(TrieError::IncompleteDatabase(<TrieHash<L>>::default())));
+			}
+		}
+
 		loop {
 			if !last_entry.advance_child_index()? {
 				last_entry.push_to_prefix(&mut prefix);
@@ -434,8 +520,14 @@ pub fn decode_compact_from_iter<'a, L, DB, T, I>(db: &mut DB, encoded: I)
 
 			// Since `advance_child_index` returned true, the preconditions for `encode_node` are
 			// satisfied.
-			let node_data = last_entry.encode_node();
-			let node_hash = db.insert(prefix.as_prefix(), node_data.as_ref());
+			let hash = last_entry.attached_value.as_ref().map(|value| {
+				db.insert(prefix.as_prefix(), value)
+			});
+			let node_data = last_entry.encode_node(hash.as_ref().map(|h| h.as_ref()));
+			let node_hash = db.insert(
+				prefix.as_prefix(),
+				node_data.as_ref(),
+			);
 
 			if let Some(entry) = stack.pop() {
 				last_entry = entry;
